/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */
package org.ginf;

import org.w3c.rdf.*;
import org.xml.sax.InputSource;

import java.util.Enumeration;

/**
 * Schema registry is used to manage run-time schema information.
 * It also defines the vocabulary used in the <a href="http://www.w3.org/TR/WD-rdf-schema">RDF schema</a> specification.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public interface SchemaRegistry {

  // Management of known schemata

  /**
   * Tests if the URI of the schema is known.
   * "Known" = do not try to fetch
   */
  public boolean isKnown(String uri);

  /**
   * Adds a schema URI to the list of known URIs.
   * If a schema URI is known no implicit attempt will be made
   * to fetch it. However, it still can be fetched on request.
   */
  public void addKnown(String uri);

  /**
   * Removes schema URI from the list of known schemas.
   */
  public void removeKnown(String uri);

  /**
   * Lists all known schemas
   */
  public Enumeration knownSchemas();
  
  // Schema management

  /**
   * Tests if the registry contains schema located at the given URI.
   */
  public boolean contains(String uri);

  /**
   * Adds the schema model to the schema registry.
   */
  public void add(String uri, Model schema);

  /**
   * Adds a new schema loader
   */
  public void addLoader(SchemaLoader loader);

  /**
   * Removes schema loader
   */
  public void removeLoader(SchemaLoader loader);

  /**
   * Tests whether contains loader
   */
  public boolean containsLoader(SchemaLoader loader);

  /**
   * Lists all registered loaders
   */
  public Enumeration loaders();

  /**
   * Removes a schema model out of the registry.
   */
  public void remove(String uri);

  /**
   * Gets the model of the schema located at URI.
   */
  public Model get(String uri);

  /**
   * Gets the complete schema information available
   * in a single model.
   */
  public Model get();

  /**
   * A callback method which is supposed to be invoked
   * by RDFMS (parser/serializer) when it
   * encounteres an unknown schema.
   */
  public void fetch(String uri);

  /**
   * A default method to load schemas from an InputSource
   * source.getSystemId() = schema URI
   */

  public void fetch(InputSource source);

}
