/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package org.ginf.helpers;

import org.w3c.rdf.*;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import org.ginf.*;
import org.ginf.schemas.RDF;
import org.ginf.schemas.RDFS;

import java.util.*;
import java.io.*;
import java.net.URL;

/**
 * A default implementation of the RDFMS interface.
 * For parsing it uses a modified version of
 * SiRPAC parser written by Janne Saarela at W3C.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public class RDFMSImpl extends SiRPAC implements RDFMS, Comparer, RDFConsumer {

  static String DEFAULT_PARSER = "com.microstar.xml.SAXDriver";

  RDFFactory factory;
  Model model;

  /**
   * the properties that require fetching schemas
   */
  static Hashtable schemaProperties;
  SchemaController controller;

  static String getParser() {

    String parser = null;
    try {
      parser = System.getProperty("org.xml.sax.parser");
    } catch( SecurityException sec) {
      // if run as an applet
    }
    return parser != null ? parser : DEFAULT_PARSER;
  }

  static {

    schemaProperties = new Hashtable();

    schemaProperties.put( RDF.type, RDF.type );
    schemaProperties.put( RDFS.isDefinedBy, RDFS.isDefinedBy );
    schemaProperties.put( RDFS.subClassOf, RDFS.subClassOf );
    schemaProperties.put( RDFS.subPropertyOf, RDFS.subPropertyOf );
    schemaProperties.put( RDFS.domain, RDFS.domain );
    schemaProperties.put( RDFS.range, RDFS.range );
  }


  public RDFMSImpl(RDFFactory factory) {

    super( getParser() );
    this.factory = factory;
  }

  /**
   * Overwrites fetchSchema method of SiRPAC
   */
  public void fetchSchema(String uri) {
    
    //    System.err.println("--- FETCH: " + uri);
    if(uri != null && factory.registry() != null)
      factory.registry().fetch(uri);
  }

  public void setSchemaController(SchemaController c) {
    controller = c;
  }

  // Implementation of the RDFConsumer interface

  /**
   * start is called when parsing of data is started
   *
   * @param	ds	DataSource starting its job
   */
  public void start (DataSource ds) {
  }
  
  /**
   * end is called when parsing of data is ended
   *
   * @param	ds	DataSource finishing its job
   */
  public void end   (DataSource ds) {
  }
  
  /**
   * assert is called every time a new statement within
   * RDF data model is added
   *
   * @param	ds	DataSource who assert'ed this statement
   *                  (can be null if the assert'ing party
   *                   does no implement DataSource interface)
   */
  public void assert (DataSource ds,
		      Resource predicate,
		      Resource subject,
		      RDFnode  object) {

    Triple t = new Triple( predicate, subject, object );

    if( controller == null && schemaProperties.containsKey(predicate) ) {
      // we are the default controller
      fetchSchema(getFetchableNS( object.toString() ));

    } else if(controller != null) {

      controller.loadSchema(t, model, factory);
      /*
      String ns = getFetchableNS( subject.toString() );
      if(controller.fetchSchema(ns, t, model))
	fetchSchema(ns);

      if(object instanceof Resource) {
	ns = getFetchableNS( object.toString() );
	if(controller.fetchSchema(ns, t, model))
	  fetchSchema(ns);
      }
      */
    }
    model.add( t );
  }

  String getFetchableNS(String url) {
    
    String base = model.getURI();

    if(base == null && url.startsWith("#"))
      return null;
    if(base != null && url.startsWith(base))
      return null;

    String ns = RDFUtil.getNS(url);
    
    try {
      URL test = new URL(ns);
      return ns;
    } catch (Exception any) {}
    
    return null;
  }

  public Model parse(InputSource source) throws Exception {

    SiRPAC compiler = this;

    compiler.setRDFSource ( source );
    compiler.fetchSchemas ( true );
    //    compiler.createBags( true );
    compiler.register (this);

    model = factory.createModel();
    model.setURI( source.getSystemId() );

    try {
      compiler.fetchRDF ();
    } catch (SAXException any) {
      //      System.err.println("INTERMEDIATE: " + any + ", " + any.getException());
      //      if(!any.getException() instanceof EndOfRDFException IOException)
      throw any.getException();
    }

    String sErrors = compiler.errors ();
    if (sErrors != null && sErrors.length() > 0)
      throw new Exception("SiRPAC Parser error: " + sErrors);

    /*

    // fetch unknown pieces
    
    String uri = model.getURI();
    
    //    Hashtable toFetch = new Hashtable(); // resource -> triple; do not override rdf:type
    
    for(Enumeration en = model.elements(); en.hasMoreElements();) {
      
      Triple t = (Triple)en.nextElement();
      
      shouldFetchSchema(uri, t.subject().toString(), t, model);
      
      if(t.object() instanceof Resource)
	shouldFetchSchema(uri, t.object().toString(), t, model);
    }
    */
    
    // we cannot return schema-aware model from the very beginning!
    return model;
  }

  public int compare(Object handle, Object o1, Object o2) {

    int t = ((Integer)handle).intValue();
    Triple t1 = (Triple)o1;
    Triple t2 = (Triple)o2;

    int res = t1.subject().toString().compareTo( t2.subject().toString() );

    if(res != 0)
      return res;

    /*
      // move rdf:type upwards
    boolean r1 = RDF_TYPE.equals( t1.predicate() );
    boolean r2 = RDF_TYPE.equals( t2.predicate() );
    
    if(r1 && !r2)
    return -1;
      if(r2 && !r1)
    return 1;
      */

    // sort abbreviatable objects before long objects
    RDFnode n1 = t1.object();
    RDFnode n2 = t2.object();

    // which is a resource?
    boolean r1 = n1 instanceof Resource;
    boolean r2 = n2 instanceof Resource;

    // push resources down
    if(r1 && r2)
      return n1.toString().compareTo( n2.toString() );
    if(r2)
      return -1;
    else if(r1)
      return 1;

    // both are literals
    // which can be abbreviated?
    r1 = canAbbrev( n1.toString() );
    r2 = canAbbrev( n2.toString() );

    if(r1 && !r2)
      return -1;
    else if(r2 && !r1)
      return 1;

    // in case of literals, first sort them out
    // according to their names (important for postponed elements)
    res = t1.predicate().toString().compareTo( t2.predicate().toString() );
    if(res != 0)
      return res;
    else // compare their values
      return n1.toString().compareTo( n2.toString() );
  }

  final char ABB_LONG = (char)0;
  final char ABB_CDATA = (char)1;

  final char ANYQUOTE = (char)0;
  final int MAX_ABBLENGTH = 60;

  /**
   * @return Quote sigh that can be used to abbreviate this string 
   * as an attribute. (char)0 if is CDATA or too long.
   */
  char abbrevQuote( String s ) {

    char quote = ANYQUOTE; // any
    boolean hasBreaks = false;
    boolean whiteSpaceOnly = true;

    for(int i=0; i < s.length(); i++) {
      char c = s.charAt(i);
      if(c == '<' /*|| c == '>'*/ || c == '&')
	return ABB_CDATA;
      else if(c == '\n')
	hasBreaks = true;

      if(c == '"' || c == '\'') {
	if(quote == ANYQUOTE)
	  quote = (c == '"') ? '\'' : '"';
	else if (c == quote)
 	  return ABB_CDATA;
      }

      if(!Character.isWhitespace(c))
	whiteSpaceOnly = false;
    }

    if(whiteSpaceOnly && hasBreaks)
      return ABB_CDATA;

    if(hasBreaks || s.length() > MAX_ABBLENGTH) // optically nice value
      return whiteSpaceOnly ? ABB_CDATA : ABB_LONG;

    return quote == ANYQUOTE ? '"' : quote;

  }

  boolean canAbbrev(RDFnode n) {
    return (n instanceof Literal) && canAbbrev(n.toString());
  }

  boolean canAbbrev(String s) {

    // just to speed up for long strings
    if(s.length() > MAX_ABBLENGTH)
      return false;

    char c = abbrevQuote(s);
    return c == '"' || c == '\'';
  }

  Hashtable getNamespaces(Enumeration en) {

    Hashtable h = new Hashtable();

    for(; en.hasMoreElements(); ) {
      Triple t = (Triple)en.nextElement();
      /*
      getNamespace(h, t.subject().getURI());
      RDFnode n = t.object();
      if(n instanceof Resource)
	getNamespace(h, ((Resource)n).getURI());
	*/
      // collect types
      if(RDF.type.equals(t.predicate())) // RDFMS.type
	getNamespace(h, t.object().toString());

      getNamespace(h, t.predicate().toString());
    }
    return h;
  }

  void getNamespace(Hashtable h, String uri) {
 
    h.put(RDFUtil.getNS(uri), uri);
    //    System.out.println("INSERTING: " + RDFUtil.getNS(uri) + " for " + uri);
  }


  final String INDENT_ABB = "\t ";
  final String INDENT_LONG = "\t";


  public void serialize(Model m, Writer w) throws IOException {

    if(m == null)
      return;

    model = m;

    PrintWriter out = new PrintWriter(w);
    out.println("<?xml version='1.0' encoding='ISO-8859-1'?>");

    if(m.size() == 0) {
      // write empty model
      out.println("<rdf:RDF xmlns:rdf=\"" + RDF.NS + "\"/>");
      out.flush();
      return;
    }

    //    String modelURI = m.getURI();

    // triples that have been written out
    Hashtable done = new Hashtable();

    Vector tv = RDFUtil.getTripleVector(m);
    // sort order: GROUP BY resource, abbreviatable
    QuickSort.sort(tv, this, new Integer(1));

    out.print("<rdf:RDF ");

    // collect namespaces

    Hashtable ns = getNamespaces(tv.elements());
    int counter = 0;
    String indent = "";
    Hashtable namespaces = new Hashtable();
    // just to be sure
    ns.put(RDF.NS, "rdf");

    for(Enumeration en = ns.keys(); en.hasMoreElements();) {

      // assign a shortcut to a namespace

      String s = (String)en.nextElement();

      //      System.out.println("URL: " + modelURI + ", comp " + s);

      String shortcut;
      if(RDF.NS.equals(s))
	shortcut = "rdf";
      else if(RDFS.NS.equals(s))
	shortcut = "s";
      //      else if(modelURI != null && modelURI.equals(s))
      //	shortcut = "this";
      else
	shortcut = counter < (int)'s' - 1 ? String.valueOf((char)((int)'a' + counter++)) : "n" + (++counter - (int)'s' + 1);

      namespaces.put(s, shortcut);
      out.print(indent + "xmlns:" + shortcut + "=\"" + s + "#\"");

      indent = "\n" + INDENT_ABB;
    }
    out.println(">");


    // contains triples having the same subject
    Vector group = new Vector();

    Triple tCurrent = null, tNext = null;
    StringBuffer buf = new StringBuffer();

    // loop over all triples
    for(int i=0; i <= tv.size(); i++) { // loop one more time than needed

      tCurrent = tNext;
      tNext = null;

      if(i < tv.size()) {
	tNext = (Triple)tv.elementAt(i);
      }

      // if different resource starts
      if(tCurrent == null || tNext == null || !tCurrent.subject().equals(tNext.subject()) ) {

	if(i > 0) {
	  // process group of triples with equals subjects, contains at least 1 element

	  //	  System.err.println("@" + i + " STARTING GROUP: " + tCurrent.subject() + ", " + group.size());

	  // Problem 1: find the first "type" and remove it from the group
	  Triple type = null;
	  for(int j=0; j < group.size(); j++) {
	    Triple tEl = (Triple)group.elementAt(j);
	    if( tEl.predicate().equals(RDF.type) ) {
	      type = tEl;
	      group.removeElementAt(j);
	    }
	  }

	  // Problem 2: find properties that can be abbreviated and have equal names
	  // Move all of them to the "bottom"
	  int abbNum = 0; // number of abbreviatable
	  int endOfGroup = group.size();

	  for(int j=0; j < endOfGroup; j++) {
	    
	    Triple tEl = (Triple)group.elementAt(j);
	    
	    //	    System.err.println("TESTING PRED: " + abbNum + ", " + endOfGroup + ", " + tEl.predicate());
	  
	    if(!canAbbrev(tEl.object())) {
	      abbNum = j;
	      break;

	    } else {

	      abbNum++;
	    
	      if(j > 0 && tEl.predicate().equals( ((Triple)group.elementAt(j-1)).predicate() )) {
		// move
		group.removeElementAt(j);
		group.addElement(tEl);
		endOfGroup--;
		j--;
	      }
	    }
	  }

	  // prepare variables
	  Resource subj = tCurrent.subject();
	  String qualName = null; // qualified name of this resource (if any)
	  if(type != null)
	    qualName = namespaces.get( RDFUtil.getNS(type.object().toString()) ) + ":" +
	      RDFUtil.getName(type.object().toString());
	  boolean shortDescription = abbNum == group.size();

	  // now: "type" contains type of the described resource (+ qualName)
	  //      abbNum contains the number of abbreviatable resources
	  //      subj   contains the subject of the triple
	  //      shortDescription ( /> )

	  //	  System.err.println("PREPARE VARS: " + abbNum);
	  
	  // start description
	  if(type != null)
	    out.print("<" + qualName);
	  else
	    out.print("<rdf:Description");
	  out.print(IDorAbout(subj.toString()));

	  // process abbreviated literals
	  //	  indent = " ";
	  for(int j=0; j < abbNum; j++) {
	    
	    Triple tEl = (Triple)group.elementAt(j);
	    Resource pred = tEl.predicate();
	    RDFnode obj = tEl.object();
	    // subj, pred, obj
	    
	    String p_ns = (String) namespaces.get( RDFUtil.getNS(pred.toString()) );
	    String p_name = RDFUtil.getName(pred.toString());
	    char quote = abbrevQuote( obj.toString() );
	    
	    out.print(indent + p_ns + ":" + p_name + "=" + quote + obj + quote);
	    //	    indent = "\n" + INDENT_ABB;
	  }

	  if(shortDescription)
	    // we are done
	    out.println("/>");
	  else {

	    out.println('>');
	    // process resources and long literals (can also be postponed)

	    for(int j=abbNum; j < group.size(); j++) {
		
	      Triple tEl = (Triple)group.elementAt(j);
	      Resource pred = tEl.predicate();
	      RDFnode obj = tEl.object();

	      String p_ns = (String) namespaces.get( RDFUtil.getNS(pred.toString()) );
	      String p_name = RDFUtil.getName(pred.toString());

	      if(obj instanceof Resource) {
		// resource
		out.println(INDENT_LONG + "<" + p_ns + ":" + p_name + " rdf:resource=\"" +
			    escapeAttValue( (RDFUtil.isQualified(obj.toString()) ? resourceID( obj.toString() ) : "#" + obj) ) +
			    "\"/>");
	      } else {
		
		char quote = abbrevQuote( obj.toString() );
		
		out.print(INDENT_LONG + "<" + p_ns + ":" + p_name);
		if(quote == ABB_CDATA)
		  out.print(" xml:space='preserve'");
		out.print('>');

		if(quote == ABB_CDATA) {
		  out.print("<![CDATA[");
		  // write out CDATA
		  escapeCDATA(out, obj.toString());
		} else
		  out.print(obj.toString());
		
		if(quote == ABB_CDATA)
		  out.print("]]>");
		out.println("</" + p_ns + ":" + p_name + ">");
	      }
	    }

	    // end description
	    if(type != null)
	      out.println("</" + qualName + ">");
	    else
	      out.println("</rdf:Description>");
	  }
	  
	  // end processing of the group
	  group.setSize(0);
	}

      }

      if(tNext != null) // continue building group
	group.addElement( tNext );

    } // loop over all triples

    out.println("</rdf:RDF>");
    out.flush();
    /*
    for(int i=0; i < tv.size(); i++) {
      Triple t = (Triple)tv.elementAt(i);
      w.write ("triple(\""+t.predicate()+"\",\""+t.subject()+"\",\""+t.object()+"\").\n");
    }
    */
  }

  /**
   * ]]> cannot be there!
   * FIXME: how to we encode binary data? It is a function
   * of some more abstract layer?
   */
  void escapeCDATA(PrintWriter out, String str) {

    int start = 0, i = 0;
    do {
      i = str.indexOf("]]>", start);
      if(i >= 0) {
	out.print(str.substring(start, i+2));
	out.print("]]><![CDATA[");
	start = i + 2;
      } else
	out.print(str.substring(start));
    } while (i >= 0 && start < str.length());
  }
  /**
   * < and &
   */
  void escapeAttValue(StringBuffer buf, String str) {

    for(int i=0; i < str.length(); i++) {
      char c = str.charAt(i);
      if(c == '<')
	buf.append("&lt;");
      else if(c == '&')
	buf.append("&amp;");
      else
	buf.append(c);
    }
  }

  String escapeAttValue(String s) {

    for(int i=0; i < s.length(); i++) {
      char c = s.charAt(i);
      if(c == '<' || c == '>' || c == '&') {
	StringBuffer buf = new StringBuffer();
	escapeAttValue(buf, s);
	return buf.toString();
      }
    }
    return s;
  }

  String IDorAbout(String s) {

    //    System.err.println("ID or ABOUT: " + model.getURI() + ", " + s + ", " + RDFUtil.getNS(s));

    if((model.getURI() != null && s.startsWith(model.getURI())) ||
       (model.getURI() == null && RDFUtil.getNS(s) == null) ) {
      return " rdf:ID=\"" + RDFUtil.getName(s) + "\"";
    } else
      return " rdf:about=\"" + escapeAttValue(s) + "\"";
  }

  String resourceID(String s) {

    if((model.getURI() != null && s.startsWith(model.getURI())) ||
       (model.getURI() == null && RDFUtil.getNS(s) == null) ) {
      return "#" + RDFUtil.getName(s);
    } else
      return s;
  }

  /**
   * main method for running DefaultRDFMS as an application
   */
  /*
  public static void main(String args[]) throws Exception {

    if (args.length == 0) {
      System.err.println("Usage: java -Dorg.xml.sax.parser=<classname> org.ginf.helpers.RDFMSImpl [ URI | filename ]");
      //	    System.err.println ("This is revision "+REVISION);
      System.exit(1);
    }

    URL url = null;
    try {
      url = new URL (args[0]);
    } catch (Exception e) {
      url = new URL ("file", null, args[0]);
    }

    RDFFactory factory = new RDFFactoryImpl();

    //    factory.registry().addKnown( "http://purl.org/metadata/dublin_core#" );
    RDFMS p = factory.createRDFMS();

    //    InputSource source = new InputSource( new InputFilter ( url.openStream() ) );
    //    InputSource source = new InputSource( new DoNotCloseFilter ( url.openStream() ) );
    InputSource source = new InputSource(  url.openStream() );
    source.setSystemId( url.toString() );
    //    System.out.println("SYS ID: " + url.toString() );
    //source.setSystemId( "http://www.w3.org/1999/02/22-rdf-syntax-ns" );

    Model m = p.parse( source );

    RDFUtil.printTriples(m, System.out);
    p.serialize( m, new OutputStreamWriter(System.out));

    m = factory.createSchemaModel( m );

    //    System.err.println("Full schema info:");
    //    DefaultModel.printTriples(registry.get(), System.err);

    System.err.println("VALIDATION:");
    ((SchemaModel)m).validate();
    
  }
  */
}
