/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package org.ginf.helpers;

import org.xml.sax.InputSource;
import org.w3c.rdf.*;
import org.ginf.*;
import org.ginf.schemas.RDF;
import org.ginf.schemas.RDFS;

import java.util.*;
import java.net.*;
import java.io.*;

/**
 * Default schema registry implementation.
 * 
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public class SchemaRegistryImpl implements SchemaRegistry, SchemaLoader {

  RDFFactory factory;
  /**
   * Stores to each schema URI its Model
   *
   * @see SM.rdf.Model
   */
  Hashtable schemata;
  
  /**
   * List of schematas which are implicitly known by
   * the application and do not have to be loaded.
   */
  Hashtable known_schemata;

  /**
   * Ordered list of SchemaLoader's.
   * Last registered - first asked
   */
  Vector loaders = new Vector();

  /**
   * Holds the complete schema information available.
   */
  Model complete;

  Model closure;
  boolean validClosure = false;

  /**
   * Uses model m as a factory to create a new model
   * to hold complete schema information.
   * This model must be flat i.e. is not allowed to
   * resolve recursive dependencies like subclassing etc.
   */
  public SchemaRegistryImpl(RDFFactory factory) /*throws Exception*/ {

    this.factory = factory;
    schemata = new Hashtable();
    known_schemata = new Hashtable();
    
    complete = factory.createModel();
    closure = factory.createModel();
    
    // I'm the default loader
    addLoader(this);
    addLoader(new RDF());
    addLoader(new RDFS());
  }

  /**
   * Lists all known schemas
   */
  public Enumeration knownSchemas() {

    return known_schemata.elements();
  }

  public void addKnown(String uri) {

    known_schemata.put(uri, uri);
  }

  public void removeKnown(String uri) {

    known_schemata.remove(uri);
  }

  public boolean isKnown(String uri) {

    return known_schemata.contains(uri);
  }

  public boolean contains(String uri) {

    return schemata.containsKey(uri);
  }

  public void add(String uri, Model schema) {

    if(contains(uri))
      return;

    schemata.put(uri, schema);
    // append the new schema to the complete schema
    complete.union( schema );
    validClosure = false;
    System.err.println("Registred schema " + uri + " of size " + schema.size());
  }

  public void remove(String uri) {

    // FIXME: remove triples from complete, recompute closure
    schemata.remove(uri);
  }

  public Model get(String uri) {

    return (Model)schemata.get(uri);
  }

  public Model get() {

    // duplicate it to make sure noone
    // occasionaly modifies schema info
    return complete.duplicate();
  }

  public void fetch(InputSource source) {

    if(source != null) {
      
      String uri = source.getSystemId();
      // to prevent dead recursion
      addKnown( uri );

      try {

	RDFMS p = factory.createRDFMS();

	Model model = p.parse( source );
	//	System.err.println("Fetched schema " + uri + " of size " + model.size());
	add(uri, model);

      } catch (Exception any) {

 	System.err.println("Schema " + uri + " is corrupted: " + any);
	remove(uri);
      }
    }
  }

  public void fetch(String uri) {

    // remove hash at the end of the uri
    //    String sysID;
    if(uri.endsWith("#"))
      uri = uri.substring(0, uri.length()-1);
    //    else
    //      sysID = uri;

    //    System.out.println("FETCHING SCHEMA: " + this + " " + uri);

    if(!isKnown(uri) &&
       !contains(uri)) {

      // is any of registered loaders able to load the schema
      for(Enumeration en = loaders(); en.hasMoreElements();) {

	SchemaLoader loader = (SchemaLoader)en.nextElement();
	if(loader.loadSchema(uri, factory))
	  return; // done
      }
    } else {
      //      System.err.println("Schema " + uri + " is known or already fetched.");
    }

  }

  /**
   * Returns a flat model in which
   * subClassOf and subPropertyOf are transitively closed
   */
  public synchronized Model getClosure() {
    
    if(!validClosure) {
      // System.err.println("COMPUTING CLOSURE");
      computeClosure( RDFS.subClassOf );
      computeClosure( RDFS.subPropertyOf );
      //      System.err.println("CLOSURE DONE");
      validClosure = true;
    }
    return closure;
  }

  void computeClosure(Resource predicate) {
      
    Model base = closure;
    base.union( get().find( null, predicate, null ) );
    Model all = base;

    int old_size;
    do {
      Model added = factory.createModel();
	
      for(Enumeration en_p = base.elements(); en_p.hasMoreElements();) {
	Triple pivot = (Triple)en_p.nextElement();
	
	for(Enumeration en_a = all.elements(); en_a.hasMoreElements();) {
	  Triple t = (Triple)en_a.nextElement();
	  if(t.subject().equals(pivot.object()))
	    added.add( pivot.subject(), predicate, t.object() );
	  if(pivot.subject().equals(t.object()))
	    added.add( t.subject(), predicate, pivot.object() );
	}
      }
      old_size = base.size();
      base.union(added);
    } while (base.size() > old_size);

    closure.union( base );
  }


  /**
   * Adds a new schema loader
   */
  public void addLoader(SchemaLoader loader) {

    loaders.insertElementAt(loader, 0);
    //    System.err.println("--- ADDED LOADER: " + loaders.size() + " " + loader);
  }

  /**
   * Removes schema loader
   */
  public void removeLoader(SchemaLoader loader) {

    loaders.removeElement(loader);
  }

  /**
   * Tests whether contains loader
   */
  public boolean containsLoader(SchemaLoader loader) {
    
    return loaders.contains(loader);
  }

  /**
   * Lists all registered loaders
   */
  public Enumeration loaders() {
    
    return loaders.elements();
  }


  // SchemaLoader implementation

  public boolean loadSchema(String uri, RDFFactory factory) {

    // prevents fetching unavailable schemas again and again
    addKnown( uri );

    try {
	
      System.err.println("CONNECTING: " + uri);
      InputSource source = new InputSource( new URL(uri).openStream() );
      source.setSystemId(uri);
      fetch(source);
      
    } catch (IOException any) {
      
      System.err.println("Schema " + uri + " is unavailable: " + any);
      remove(uri);
    }

    return true; // catch everything
  }

}
